///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// File:	hr$dirtool.js
//
// Purpose:	General purpose functions for scanning and listing folder contents.
//
// Author:	Copyright (c) 2005 Henk Reints, http://henk-reints.nl
//
// Prerequisites:
//		For manipulating File and Folder Objects, a FileSystemObject Object should be available,
//		which is expected to be accessible via a (global) variable named "FSO".
//		Some functions work only if used with the Cscript Windows Script Host.
//
// Update history:
//	2007-07-12	fixed a stupid bug in the sorting algorithm, see below
//	2007-07-05	use my new Date.parseFileDate method defined in hr$wsh-fso.js
//	Nov/Dec 2005	original version
//
// DISCLAIMER:
//	HENK REINTS HAS WRITTEN THIS SOFWARE FOR HIS OWN USE, AS A HOBBY, WITHOUT ANY COMMERCIAL INTENTIONS.
//	IT IS PROVIDED AS IS, WITHOUT ANY GUARANTEE FOR CORRECTNESS OR USABILITY. USAGE IS AT YOUR OWN RISK.
//	IN NO WAY SHALL HENK REINTS BE LIABLE FOR ANY NEGATIVE CONSEQUENCES OF USING THIS SOFTWARE.
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	hr$dirtool = new Object()

// ======================================
// Properties to be considered Read Only:
// --------------------------------------
										// ###########################
	hr$dirtool.version = "2007.07.05  Copyright (c) Henk Reints"		// KEEP VERSION UP-TO-DATE !!!
										// ###########################
	hr$dirtool.FolderNameField		= 1
	hr$dirtool.FileNameField		= 2
	hr$dirtool.FileTypeField		= 3
	hr$dirtool.DateLastModifiedField	= 4
	hr$dirtool.DateCreatedField		= 5
	hr$dirtool.DateLastAccessedField	= 6
	hr$dirtool.SizeField			= 7
	hr$dirtool.AttributesField		= 8
	hr$dirtool.ExtraInfoField		= 9

	hr$dirtool.attrReadOnly		=   1
	hr$dirtool.attrHidden		=   2
	hr$dirtool.attrSystem		=   4
	hr$dirtool.attrVolumeID		=   8
	hr$dirtool.attrDirectory	=  16
	hr$dirtool.attrArchive		=  32
	hr$dirtool.attrLink		=  64
	hr$dirtool.attrCompressed	= 128
	hr$dirtool.attrCharacters	= "rhsvdalc"

// ======================
// Read/Write properties:
// ----------------------

	hr$dirtool.sortBufferSize	= 2048
	hr$dirtool.tempfile		= ""	// set at the end of this file, when all methods have been defined
	hr$dirtool.verbose		= false
	hr$dirtool.useExtraInfo		= false	// _IF_ you have defined a hr$dirtool.getExtraFileFolderInfo method
						// then this property tells whether to use it or not.

// ====================================================================================================================
// Method "scanFolder" is a general folder scanner.
//
// Arguments:
//	folderObject		Folder Object of the folder to be scanned.
//
//	fileFolderObjectHandler	Function to handle files and subfolders found in the scanned folder.
//				Invoked once for each file or folder found.
//				It should have arguments: (fileFolderobject, isFolder)
//				where  fileFolderobject  is the File or Folder Object
//				and    isFolder          is a boolean indicating if it is a folder.
//
//	recursive		Boolean value indicating whether to scan the entire subfolder tree as well.
//
// Returns:
//	<nothing>
//
// Special behaviour:
//	You can redefine the hr$dirtool.fileFilter method and/or the folderFilter method.
//	They receive the File or Folder Object as argument and should return true or false to indicate if the file or
//	folder should be handled or not. If the filter returns 'false' then the file or folder will be ignored, and
//	in case of a subfolder it will not be recursed into. It could for example handle an exclude list of files
//	or folders, or filter by date, size, attributes or whatsoever.
//	Example (which will exclude all folders named "temp" and all their subfolder trees):
//
//		hr$dirtool.folderFilter = function (folderObject)
//		{	return (FSO.GetFileName(folderObject.Path).toLowercase() != "temp")
//		}
// --------------------------------------------------------------------------------------------------------------------

	hr$dirtool.passAll      = function() {return true}	// default filter to include everything
	hr$dirtool.fileFilter   = hr$dirtool.passAll		// you can redefine this to filter files out
	hr$dirtool.folderFilter = hr$dirtool.passAll		// you can redefine this to filter folders out

	hr$dirtool.scanFolder = function (folderObject, fileFolderObjectHandler, recursive)
	{
		var fileCollection = new Enumerator(folderObject.Files)
		for (; !fileCollection.atEnd(); fileCollection.moveNext() )
		{	fileObject = fileCollection.item()
			if (hr$dirtool.fileFilter (fileObject) ) fileFolderObjectHandler (fileObject, false)
		}
		var subFolderCollection = new Enumerator(folderObject.SubFolders)
		for (; !subFolderCollection.atEnd(); subFolderCollection.moveNext() )
		{	var subFolderObject = subFolderCollection.item()
			if (hr$dirtool.folderFilter (subFolderObject) )
			{	fileFolderObjectHandler (subFolderObject, true)
				if (recursive) hr$dirtool.scanFolder (subFolderObject, fileFolderObjectHandler, true)
			}
		}
	}

// ====================================================================================================================
// Method "getFileFolderList" obtains a full listing of a folder or folder tree.
//
// Arguments:
//	folderObject		Folder Object of the folder to be scanned.
//
//	recursive		Boolean value indicating whether to scan the entire subfolder tree as well.
//
//	wildcardPattern		Pattern against which file names are checked.
//				Only matching file names are included in the result.
//				if the pattern is prefixed with a slash then it is an exclusion filter.
//
//	onlyFiles		Boolean, if true then folders are not included in the result.
//
//	onlyFolders		Boolean, if true then files are not included in the result.
//
//	propertyFilter		Optional argument, Array with extra filters
//				[0] = inclusive date filter, include file if (start of) filedate matches
//				[1] = exclusive date filter, exclude file if (start of) filedate matches
//				[2] = since-date
//				[3] = before-date
//				[4] = minsize
//				[5] = maxsize
//				[6] = inclusive attributes mask for boolean OR
//				[7] = exclusive attributes mask for boolean OR
//				[8] = inclusive attributes mask for boolean AND
//				[9] = exclusive attributes mask for boolean AND
//				[10] = boolean value of onlyMultiNames
//					if true then list will be sortFileFolderList will sort by file name and
//					showFileFolderList will list only files with a name that occurs more than once
//				[11] = same for filtering multiply occurring extra info
//
//				see also help module of 'dirCommand' method for more details
//				to disable filters:
//				- date filter:	a blank string
//				- minsize:	Number.NEGATIVE_INFINITY
//				- maxsize:	Number.POSITIVE_INFINITY
//				- attr filter:	NaN
//
// Returns:
//	Array object with the information.
//	Each element is itself an Array containing the fields.
//	The returned Array has some extra properties:
//	- timestamp           = Date object with date of invocation
//	- driveObject         = the Drive Object of the root folder
//	- rootPath            = the path of the root folder, all other info contains relative paths
//	- onlyFiles           = a copy of the 'onlyFiles' argument
//	- onlyFolders         = a copy of the 'onlyFolders' argument
//	- recursive           = a copy of the 'recursive' argument
//	- filter              = the wildcard pattern actually used
//	- maxFileNameLength   = length of the longest file name
//	- maxFolderNameLength = length of the longest folder name
//	- maxSize             = greatest file size
//	- totalSize           = the total size of all files listed
//	- fileCount           = the number of files listed
//	- folderCount         = the number of folders listed
//	- totalCount          = the total size of the resulting list
//	- comment             = a blank string, you can fill it later with any text you like
//	                        (e.g. info about the propertyFilter or so)
//	- tempfile            = full path of temporary file for storage of intermediate data
//	- index               = a zero-length Array for use by the sortFileFolderList method of this package
//	- onlyMultiNames      = copy of propertyFilter[10]
//	- onlyMultiExtra      = copy of propertyFilter[11]
//
// Special behaviour:
//	You can add a method to the hr$dirtool Object called 'getExtraFileFolderInfo'.
//	If that exists, then it will be invoked for every file or folder with the File or Folder Object as argument.
//	It can then return any kind of extra information, for example something based on the file's content or so.
//	If that function is called without arguments it should return a header line to be displayed above the extra
//	info, which will be used by the showFileFolderList method (below).
//	Please note that such a function may change the file's Last Access Date.
//	The value handled here is the Last Access Date prior to invocation of 'getExtraFileFolderInfo'.
// --------------------------------------------------------------------------------------------------------------------

	hr$dirtool.getFileFolderList =
	function (folderObject, recursive, wildcardPattern, onlyFiles, onlyFolders, propertyFilter)
	{
		var N = 0, IN = 0, bs = String.fromCharCode(8), erase = 0
		var result = []
		result.timestamp           = new Date()
		result.driveObject         = folderObject.Drive
		result.rootPath            = folderObject.Path
		result.onlyFiles           = onlyFiles
		result.onlyFolders         = onlyFolders
		result.recursive           = recursive
		result.filter              = ""
		result.maxFileNameLength   = 0
		result.maxFolderNameLength = 0
		result.maxSize             = 0
		result.totalSize           = 0
		result.fileCount           = 0
		result.folderCount         = 0
		result.totalCount          = 0
		result.comment             = ""
		result.tempfile            = ""
		result.index               = []
		result.onlyMultiNames      = false
		result.onlyMultiExtra      = false

		var tempfile
		var inclDate = "", exclDate = "", since = "", before = ""
		var minsize = Number.NEGATIVE_INFINITY
		var maxsize = Number.POSITIVE_INFINITY
		var aORi = NaN, aORe = NaN, aANDi = NaN, aANDe = NaN

		if (arguments.length > 5)
		{	inclDate		= propertyFilter[0]
			exclDate		= propertyFilter[1]
			since			= propertyFilter[2]
			before			= propertyFilter[3]
			minsize			= propertyFilter[4]
			maxsize			= propertyFilter[5]
			aORi			= propertyFilter[6]
			aORe			= propertyFilter[7]
			aANDi			= propertyFilter[8]
			aANDe			= propertyFilter[9]
			result.onlyMultiNames	= propertyFilter[10]
			result.onlyMultiExtra	= propertyFilter[11]
		}
		if (before == "") before = "~~~~~~~~~~~~~~"

		var root = result.rootPath; if (root.slice(-1) != "\\") root += "\\"

		var filter = wildcardPattern
		var exclude = filter.charAt(0) == "/"
		if (exclude) filter = filter.slice(1)
		if (filter.indexOf(".") < 0) filter += ".*"
		if (filter.charAt(0) == ".") filter = "*" + filter

		result.filter = (exclude ? "/" : "") + filter

		filter = filter.replace(/([\$\(\)\+\.\[\\\^\{\|])/g,"\\$1").replace(/\?/g,".").replace(/\*/g,".*?")
		var filterRE = new RegExp ('^'+filter+'$', 'i')

		function storeFileFolderInfo (fileFolderObject, isFolder)
		{	IN++
			if (hr$dirtool.verbose && (IN%1000 == 0 || (IN < 1000 && IN%100 == 0) ) )
			{	var eratxt = (new Array(erase+1)).join(bs)
				var logtxt = " "+IN
				erase = logtxt.length
				WScript.StdErr.Write(eratxt+logtxt)
			}
			var path = fileFolderObject.Path
			var name = FSO.GetFileName (path)
			var test = name
			if (test.indexOf(".") < 0) test += "."
			var match = (test.search(filterRE) >= 0)
			if ( (match && exclude) || (!match && !exclude) ) return

			var dmod = Date.parseFileDate(fileFolderObject.DateLastModified)

			var filedate = hr$dirtool.fmtDate0(dmod)
			var dateOK = true
			if (inclDate != "" && filedate.indexOf(inclDate) != 0) dateOK = false
			if (exclDate != "" && filedate.indexOf(exclDate) == 0) dateOK = false
			if (since < before)	{if (filedate < since || filedate >= before) dateOK = false}
			else			{if (filedate < since && filedate >= before) dateOK = false}
			if (!dateOK) return

			var attr = fileFolderObject.Attributes
			var attrOK = true
			if (!isNaN(aORi ) && aORi  != 0 && (attr&aORi ) == 0    ) attrOK = false
			if (!isNaN(aORe ) && aORe  != 0 && (attr&aORe ) != 0    ) attrOK = false
			if (!isNaN(aANDi) && aANDi != 0 && (attr&aANDi) != aANDi) attrOK = false
			if (!isNaN(aANDe) && aANDe != 0 && (attr&aANDe) == aANDe) attrOK = false
			if (!attrOK) return

			if (isFolder)
			{	if (onlyFiles) return
				var size = NaN
				if (onlyFolders && !recursive)
				{	var e
					try {size = fileFolderObject.Size} catch(e) {size = NaN}
				}
			} else
			{	if (onlyFolders) return
				var size = fileFolderObject.Size
			}

			if (!isNaN(size))
			{
				var sizeOK = true
				if (minsize <= maxsize)	{if (size < minsize || size > maxsize) sizeOK = false}
				else			{if (size < minsize && size > maxsize) sizeOK = false}
				if (!sizeOK) return

				result.totalSize += size
				result.maxSize = Math.max (size, result.maxSize)
			}

			if (isFolder)	result.folderCount++
			else		result.fileCount++

			var fldr = FSO.GetParentFolderName(path).slice(root.length)
			var ftyp = FSO.GetExtensionName(path)
			var dcre = Date.parseFileDate(fileFolderObject.DateCreated)
			var dacc = Date.parseFileDate(fileFolderObject.DateLastAccessed)

			if (hr$dirtool.useExtraInfo)
			{	var e
				try {	var xtra = hr$dirtool.getExtraFileFolderInfo (fileFolderObject)
				} catch(e)
				{	hr$dirtool.useExtraInfo = false
					result.onlyMultiExtra   = false
				}
			}

			if (result.tempfile != "")
			{	N++
				if (hr$dirtool.useExtraInfo)
					tempStream.WriteLine([fldr,name,ftyp,dmod,dcre,dacc,size,attr,xtra].join("|"))
				else	tempStream.WriteLine([fldr,name,ftyp,dmod,dcre,dacc,size,attr].join("|"))
			} else
			{	if (hr$dirtool.useExtraInfo)
					result[N++] = [fldr,name,ftyp,dmod,dcre,dacc,size,attr,xtra]
				else	result[N++] = [fldr,name,ftyp,dmod,dcre,dacc,size,attr]
			}

			result.maxFileNameLength   = Math.max (name.length, result.maxFileNameLength)
			result.maxFolderNameLength = Math.max (fldr.length, result.maxFolderNameLength)
		}

		if (hr$dirtool.tempfile != "")
		{	var e, tempfile = FSO.GetAbsolutePathName(hr$dirtool.tempfile)
			try {tempStream = FSO.CreateTextFile(tempfile)}
			catch(e)
			{	WScript.StdErr.WriteLine(
					"# Error: Unable to create temp.file "+tempfile+" ["+e.description+"]")
				WScript.Quit(1)
			}
			if (hr$dirtool.verbose) WScript.StdErr.WriteLine ("# Using temp.file " + tempfile)
			result.tempfile = tempfile
		}
		if (hr$dirtool.verbose)
		{ WScript.StdErr.Write ("# Scanning folder " + (recursive ? "tree " : "") + folderObject.Path + ",")
		}

		var startTime = new Date()
		hr$dirtool.scanFolder (folderObject, storeFileFolderInfo, recursive)
		var elapsed = (new Date()).valueOf() - startTime.valueOf()

		if (hr$dirtool.verbose)
		{	var eratxt = (new Array(erase+1)).join(bs)
			var logtxt = " "+IN
			erase = logtxt.length
			WScript.StdErr.WriteLine (eratxt+logtxt+" scanned, "+N+" stored, elapsed: "
								+ (elapsed/1000).toFixed(3) + " sec")
		}
		if (result.tempfile != "") tempStream.Close()
		if (hr$dirtool.verbose)
		{	WScript.StdErr.WriteLine("# Temp.file size = " +FSO.GetFile(result.tempfile).Size + " bytes")
		}

		result.totalCount = N
		return result
	}

// ====================================================================================================================
// Method "sortFileFolderList" sorts the result of getFileFolderList.
//
// Arguments:
//	fileFolderList		Array containing the result of getFileFolderList.
//
//	sortItems		Array with sort keys.
//				Negative value is sort descending.
//				for sorting on attributes this sortItems Array should have an extra property called
//				'attrSort', which itself is an Array with the attribute bits on which to sort,
//				in order of sorting precedence, one attribute per element.
//
// Returns:
//	<nothing>	(the Array is sorted in-place)
//
// Result:
//	This method does not sort the data in place, but it fills the 'index' property of the fileFolderList with an
//	Array of indices specifying the order in which the fileFolderList must be read to obtain the sorted data.
//	If a tempfile is used then this index is not filled, but the temp file is sorted in place.
// --------------------------------------------------------------------------------------------------------------------

	hr$dirtool.sortFileFolderList = function (fileFolderList, sortItems)
	{
		var N = 0, bs = String.fromCharCode(8), erase = 0

		if (fileFolderList.onlyMultiNames)
		{	var tmp = sortItems.slice()
			sortItems.length = 0
			sortItems.push(hr$dirtool.FileNameField)
			for (var i = 0; i < tmp.length; i++)
			{	if (Math.abs(tmp[i]) != hr$dirtool.FileNameField) sortItems.push(tmp[i])
			}
		}
		else if (fileFolderList.onlyMultiExtra)
		{	var tmp = sortItems.slice()
			sortItems.length = 0
			sortItems.push(hr$dirtool.ExtraInfoField)
			for (var i = 0; i < tmp.length; i++)
			{	if (Math.abs(tmp[i]) != hr$dirtool.ExtraInfoField) sortItems.push(tmp[i])
			}
		}
		else if (sortItems.length == 0) return

		function compareFileFolderInfo (info1, info2)
		{	N++
			if (hr$dirtool.verbose && (N%10000 == 0 || (N < 10000 && N%1000 == 0) ) )
			{	var eratxt = (new Array(erase+1)).join(bs)
				var logtxt = " "+N
				erase = logtxt.length
				WScript.StdErr.Write(eratxt+logtxt)
			}
			for (var i = 0, n = 0; n == 0 && i < sortItems.length; i++)
			{	var k = sortItems[i], m = Math.abs(k), j = m - 1
				if (m == hr$dirtool.AttributesField)
				{	var i1 = info1[j], i2 = info2[j]
					for (var ii = 0; n == 0 && ii < sortItems.attrSort.length; ii++)
					{	var kk = sortItems.attrSort[ii], mm = Math.abs(kk)
						if	( (i1&mm) > (i2&mm) ) n = +kk
						else if	( (i1&mm) < (i2&mm) ) n = -kk
					}
					n *= k
				} else if (j < hr$dirtool.FileTypeField)
				{	var x1 = info1[j].toUpperCase()
					var x2 = info2[j].toUpperCase()
					if	(x1 > x2) n = +k
					else if	(x1 < x2) n = -k
				} else
				{	var x1 = info1[j]; if (typeof x1 == "number" && isNaN(x1)) x1 = -1
					var x2 = info2[j]; if (typeof x2 == "number" && isNaN(x2)) x2 = -1
					if	(x1 > x2) n = +k
					else if	(x1 < x2) n = -k
				}
			}
			return n
		}

	// next function is my own sorting algorithm, see hr$arraysort.js
	// I do not use the built-in sort method of the Array object, since that is not 'stable'
	// and my algorithm is just better and sometimes even faster when sorting LARGE amounts of data

		function createIndex (array, cmpfun)
		{	var N = array.length, index = [new Array(N), new Array(N)], iDst = 0
			for (var i = 0; i < N; index[iDst][i] = i++);
			for (var L = 1; L < N; L += L)
			{	for (var iSrc = iDst, iDst = 1 - iSrc, i2 = 0, i3 = 0; i2 < N; )
				{	for (var i1 = i2, L1 = Math.min(i1 + L, N),
						 i2 = L1, L2 = Math.min(i2 + L, N); i1 < L1 && i2 < L2; )
					{	if (cmpfun(array[index[iSrc][i1]],array[index[iSrc][i2]]) <= 0)
							index[iDst][i3++] = index[iSrc][i1++]
						else	index[iDst][i3++] = index[iSrc][i2++]
					}
					for (; i1 < L1; index[iDst][i3++] = index[iSrc][i1++] );
					for (; i2 < L2; index[iDst][i3++] = index[iSrc][i2++] );
			}	}
			return index[iDst]
		}

	// next is my own file sorting algorithm. brief description:
	// first it splits 'inpfil' in chunks of 'bufsiz' records and sorts each chuck using 'createIndex'
	// (above), and writes each sorted chunk to a temporary file in the same folder as specified by 'outfil'.
	// then it starts merging all sorted chunks together, ending up with one large sorted file.
	// it depends on the computer's configuration what the optimal 'bufsiz' value is.

		function sortFile (inpfil, outfil, bufsiz, cmpfun)
		{
			var k = 0, m, n = 0, buf = [], nam3 = outfil + ".hr$sort0.tmp"
			function writeBuf()
			{	var idx = createIndex(buf,cmpfun)
				var out = FSO.CreateTextFile (outfil + ".hr$sort" + (n++) + ".tmp")
				for (var i = 0; i < idx.length; out.WriteLine (buf[idx[i++]]) );
				out.Close(); buf.length = 0; k = 0
			}
			var inp = FSO.OpenTextFile (inpfil, 1)
			while (!inp.AtEndOfStream) {buf[k++] = inp.ReadLine(); if (k >= bufsiz) writeBuf() }
		// bugfix reintsh 2007-07-12, next line fixed. old was: "if (k > 0) writeBuf()", but if
		// inpfil is zero size it would fail at FSO.GetFile(nam3) near the end because nothing was done...
			writeBuf()
			inp.Close()
			for (k = 0, m = n; m > 1; m--)
			{	var nam1 = outfil + ".hr$sort" + (k++) + ".tmp", in1 = FSO.OpenTextFile(nam1,1)
				var nam2 = outfil + ".hr$sort" + (k++) + ".tmp", in2 = FSO.OpenTextFile(nam2,1)
				var nam3 = outfil + ".hr$sort" + (n++) + ".tmp", out = FSO.CreateTextFile(nam3)
				var rec1, have1 = !in1.AtEndOfStream; if (have1) rec1 = in1.ReadLine()
				var rec2, have2 = !in2.AtEndOfStream; if (have2) rec2 = in2.ReadLine()
				while (have1 || have2)
				{	if (have1 && have2)
					{	if (rec1 <= rec2)
							{out.WriteLine(rec1); have1 = false}
						else	{out.WriteLine(rec2); have2 = false}
					}
					else if (have1) {out.WriteLine(rec1); have1 = false}
					else if (have2) {out.WriteLine(rec2); have2 = false}
					if (!have1) {have1 = !in1.AtEndOfStream; if (have1) rec1 = in1.ReadLine()}
					if (!have2) {have2 = !in2.AtEndOfStream; if (have2) rec2 = in2.ReadLine()}
				}
				out.Close()
				in1.Close(); FSO.DeleteFile(nam1)
				in2.Close(); FSO.DeleteFile(nam2)

			}
			if (outfil == inpfil) FSO.DeleteFile(inpfil)
		// reintsh 2007-07-12 next might fail if the above bug weren't fixed:
			FSO.GetFile(nam3).Name = FSO.GetFileName(outfil)
			return n-1
		}

		function parseRecord (rec)
		{	var x = rec.split("|"), y = [x[0],x[1],x[2],1*x[3],1*x[4],1*x[5],1*x[6],1*x[7]]
			if (x.length > 8) y.push(x.slice(8).join("|"))
			return y
		}
		function compareFileFolderRecord (rec1, rec2)
		{	return compareFileFolderInfo (parseRecord(rec1), parseRecord(rec2) )
		}

		if (hr$dirtool.verbose)
		{	WScript.StdErr.Write ("# Sorting " + fileFolderList.totalCount + " records")
			if (fileFolderList.tempfile != "")
			{	WScript.StdErr.Write (" (buffer = " + hr$dirtool.sortBufferSize + ")")
			}
			WScript.StdErr.Write (",")
		}
		var startTime = new Date()

		n = NaN
		if (fileFolderList.tempfile == "")
			fileFolderList.index = createIndex (fileFolderList, compareFileFolderInfo)
		else	n = sortFile( fileFolderList.tempfile, fileFolderList.tempfile,
				hr$dirtool.sortBufferSize, compareFileFolderRecord)

		var elapsed = (new Date()).valueOf() - startTime.valueOf()
		if (hr$dirtool.verbose)
		{	var eratxt = (new Array(erase+1)).join(bs)
			var logtxt = " "+N
			erase = logtxt.length
			WScript.StdErr.WriteLine(eratxt + logtxt + " cmp's"
				+ (isNaN(n) ? "" : ", " + n + " tmp.files")
				+ ", elapsed: " + (elapsed/1000).toFixed(3) + " sec")
		}

		if (!fileFolderList.onlyMultiNames && !fileFolderList.onlyMultiExtra) return

	// now filter only files/folders with multiply occurring name

		if (hr$dirtool.verbose)
		{
			if (fileFolderList.onlyMultiNames) WScript.StdErr.Write("# Filtering multiple names")
			if (fileFolderList.onlyMultiExtra) WScript.StdErr.Write("# Filtering multiple info")
			erase = 0
		}
		var startTime = new Date()

		function saveItem(item)
		{	if (item[hr$dirtool.AttributesField-1] & hr$dirtool.attrDirectory)
			{	var L = item[hr$dirtool.FolderNameField-1].length
				fileFolderList.maxFolderNameLength = Math.max(fileFolderList.maxFolderNameLength,L)
				fileFolderList.folderCount++
			} else
			{	var L = item[hr$dirtool.FileNameField-1].length
				fileFolderList.maxFileNameLength = Math.max(fileFolderList.maxFileNameLength,L)
				fileFolderList.fileCount++
			}
			var size = item[hr$dirtool.SizeField-1]
			if (isFinite(size))
			{	fileFolderList.maxSize = Math.max(fileFolderList.maxSize,size)
				fileFolderList.totalSize += size
			}
			if (fileFolderList.tempfile != "")
			{	outStream.WriteLine(item.join("|"))
				fileFolderList.totalCount++
			} else	fileFolderList[fileFolderList.totalCount++] = item
		}

		fileFolderList.maxFileNameLength   = 0
		fileFolderList.maxFolderNameLength = 0
		fileFolderList.maxSize             = 0
		fileFolderList.totalSize           = 0
		fileFolderList.fileCount           = 0
		fileFolderList.folderCount         = 0
		fileFolderList.totalCount          = 0

		var lastItem = ["","","",NaN,NaN,NaN,NaN,NaN,""]
		var lastItemSaved = true // i.e. don't save initial Array

		if (fileFolderList.tempfile != "")
		{	var inpFile = fileFolderList.tempfile + ".old"
			FSO.GetFile(fileFolderList.tempfile).Name = FSO.GetFileName(inpFile)
			var inpStream = FSO.OpenTextFile   (inpFile, 1)
			var outStream = FSO.CreateTextFile (fileFolderList.tempfile)
		}
		for (N = 0; true; N++)
		{	if (fileFolderList.tempfile != "")
			{	if (inpStream.AtEndOfStream) break
				var item = parseRecord(inpStream.ReadLine())
			} else
			{	if (i >= fileFolderList.length) break
				var k = (fileFolderList.index.length != 0 ? fileFolderList.index[i] : i)
				var item = fileFolderList[k]
			}
			if (fileFolderList.onlyMultiNames)
			{	if (item[1].toLowerCase() == lastItem[1].toLowerCase() )
				{	if (!lastItemSaved) saveItem(lastItem)
					saveItem(item)
					lastItemSaved = true
				} else	lastItemSaved = false
			} else if (fileFolderList.onlyMultiExtra)
			{	if (item[8].toLowerCase() == lastItem[8].toLowerCase() )
				{	if (!lastItemSaved) saveItem(lastItem)
					saveItem(item)
					lastItemSaved = true
				} else	lastItemSaved = false
			} else	lastItemSaved = false
			lastItem = item.slice()

			if (hr$dirtool.verbose && (N%1000 == 0 || (N < 1000 && N%100 == 0) ) )
			{	var eratxt = (new Array(erase+1)).join(bs)
				var logtxt = " "+fileFolderList.totalCount+"/"+N
				erase = logtxt.length
				WScript.StdErr.Write(eratxt+logtxt)
			}
		}
		if (fileFolderList.tempfile == "")
			fileFolderList.length = fileFolderList.totalCount
		else
		{	inpStream.Close()
			outStream.Close()
			FSO.DeleteFile(inpFile)
		}

		if (hr$dirtool.verbose)
		{	var elapsed = (new Date()).valueOf() - startTime.valueOf()
			var eratxt = (new Array(erase+1)).join(bs)
			var logtxt = " "+fileFolderList.totalCount+"/"+N
			erase = logtxt.length
			WScript.StdErr.WriteLine(eratxt+logtxt+ ", elapsed: " + (elapsed/1000).toFixed(3) + " sec")
		}
	}

// ====================================================================================================================
// Method "showFileFolderList" displays the result of getFileFolderList.
//
// Arguments:
//	outputStream		An open output TextStream Object, for example WScript.StdOut.
//	fileFolderList		Array containing the result of getFileFolderList.
//	fieldList		Array with field numbers to be displayed.
//	basicOutput		List only paths and nothing more.
//
// Returns:
//	<nothing>
// --------------------------------------------------------------------------------------------------------------------

	hr$dirtool.showFileFolderList = function (outputStream, fileFolderList, fieldList, basicOutput)
	{
		function rpad (x, L) {for (var y = "" + x; y.length < L; y += " "    ); return y}
		function lpad (x, L) {for (var y = "" + x; y.length < L; y  = " " + y); return y}

		var e,i,j,k,k1,w,x,y
		var L0 = fileFolderList.maxFolderNameLength
		var L1 = Math.max(16,fileFolderList.maxFileNameLength)
		var L2 = Math.max(10,(""+fileFolderList.maxSize).length)

		var headers =				// must be sorted by ascending field number as defined above!
			[rpad("Location"     ,L0)
			,rpad("File Name"    ,L1)
			,rpad("Type"         , 4)
			,rpad("Last Modified",19)
			,rpad("Created"      ,19)
			,rpad("Last Accessed",19)
			,lpad("Size"         ,L2)
			,rpad("Attribs"      , 7)
			]
		if (hr$dirtool.useExtraInfo) try {headers.push(hr$dirtool.getExtraFileFolderInfo())} catch(e){}

		var lines = []
		for (j = 0; j < headers.length; j++)
		{	lines[j] = (new Array(headers[j].length+1)).join("-")
		}
		for (x = [], y = [], j = 0; j < fieldList.length; j++)
		{	k = fieldList[j]-1
			if (k < headers.length) {x.push(headers[k]); y.push(lines[k])}
		}
		if (!basicOutput)
		{	outputStream.WriteLine(x.join(" ").replace(/\s+$/,""))
			outputStream.WriteLine(y.join(" ").replace(/\s+$/,""))
		}
		L0 = Math.max(L0,headers[hr$dirtool.FolderNameField-1].length)
		L1 = Math.max(L1,headers[hr$dirtool.FileNameField  -1].length)
		L2 = Math.max(L2,headers[hr$dirtool.SizeField      -1].length)

		function showItem(item)
		{	var x,j,k1,k,w,path
			if (basicOutput)
			{	path = FSO.BuildPath (fileFolderList.rootPath, item[0])
				path = FSO.BuildPath (path, item[1])
				outputStream.WriteLine(path)
			} else
			{  for (x = [], j = 0; j < fieldList.length; j++)
			   {	k1 = fieldList[j], k = k1-1
				if (k1 > 0 && k < item.length) w = headers[k].length
				switch (k1)
				{
				case hr$dirtool.FolderNameField:	x.push(rpad(item[k],L0));		break
				case hr$dirtool.FileNameField:		x.push(rpad(item[k],L1));		break
				case hr$dirtool.FileTypeField:		x.push(rpad(item[k],w ));		break
				case hr$dirtool.DateLastModifiedField:	x.push(hr$dirtool.fmtDate(item[k]));	break
				case hr$dirtool.DateCreatedField:	x.push(hr$dirtool.fmtDate(item[k]));	break
				case hr$dirtool.DateLastAccessedField:	x.push(hr$dirtool.fmtDate(item[k]));	break
				case hr$dirtool.SizeField:
						var size = item[k];	x.push(lpad(isNaN(size)?"":size,L2));	break
				case hr$dirtool.AttributesField:	x.push(hr$dirtool.fmtAttr(item[k]));	break
				case hr$dirtool.ExtraInfoField:
						if (k < item.length)	x.push(rpad(item[k],w ));		break
				}
			   }
			   outputStream.WriteLine(x.join(" ").replace(/\s+$/,""))
			}
		}

		function parseRecord (rec)
		{	var x = rec.split("|"), y = [x[0],x[1],x[2],1*x[3],1*x[4],1*x[5],1*x[6],1*x[7]]
			if (x.length > 8) y.push(x.slice(8).join("|"))
			return y
		}

		var lastItem = []
		if (fileFolderList.tempfile != "")
		{	var tempStream = FSO.OpenTextFile (fileFolderList.tempfile, 1)
		}
		for (i = 0; true; i++)
		{	if (fileFolderList.tempfile != "")
			{	if (tempStream.AtEndOfStream) break
				var item = parseRecord(tempStream.ReadLine())
			} else
			{	if (i >= fileFolderList.length) break
				var k = (fileFolderList.index.length != 0 ? fileFolderList.index[i] : i)
				var item = fileFolderList[k]
			}
			if (i > 0)
			{	if (fileFolderList.onlyMultiNames && item[1] != lastItem[1]) outputStream.WriteLine("")
				if (fileFolderList.onlyMultiExtra && item[8] != lastItem[8]) outputStream.WriteLine("")
			}
			showItem(item)
			lastItem = item.slice()
		}
		if (fileFolderList.tempfile != "") tempStream.Close()
	}

// ====================================================================================================================
// Method "showFileFolderHeader" displays some header info of a getFileFolderList result.
//
// Arguments:
//	outputStream		An open output TextStream Object, for example WScript.StdOut.
//	fileFolderList		Array containing the result of getFileFolderList.
//	emptyLead		Number of leading empty lines displayed.
//	emptyTrail		Number of trailing empty lines displayed.
//
// Returns:
//	<nothing>
//
// Remark:
//	This methods supports the separate display of the file's extension, but a displayed file name field
//	always includes its extension, so displaying name and extension will show: "abc.txt" and "txt"
//	(maybe I'll fix this in the future, but for now I'll leave it as is).
// --------------------------------------------------------------------------------------------------------------------

	hr$dirtool.showFileFolderHeader = function (outputStream, fileFolderList, emptyLead, emptyTrail)
	{
		var driveTypeList = "Unknown/Removable/Fixed/Network/CD-ROM/RAM Disk".split("/")

		var net = new ActiveXObject("WScript.Network")
		var drv = fileFolderList.driveObject
		var R   = 0x100000000
		var sno = (drv.SerialNumber < 0 ? R+R+drv.SerialNumber : R+drv.SerialNumber).toString(16).slice(1)

		for (var i = 0; i < emptyLead; outputStream.WriteLine(""), i++);

		outputStream.WriteLine("hr$dirtool V" + hr$dirtool.version)
		outputStream.WriteLine ("Date/time     = " + fileFolderList.timestamp)
		outputStream.WriteLine ("User          = "
			+ (net.ComputerName != net.UserDomain ? net.UserDomain + "\\" : "")
			+ net.UserName
			)
		outputStream.WriteLine ("Computer      = " + net.ComputerName)
		outputStream.WriteLine ("Volume        = " + drv.VolumeName)
		outputStream.WriteLine ("Serial number = " + [sno.slice(0,4),sno.slice(4)].join("-").toUpperCase())
		outputStream.WriteLine ("File system   = " + drv.FileSystem)
		outputStream.WriteLine ("Drive type    = " + driveTypeList[drv.DriveType])

		if (drv.Sharename != "" && drv.Sharename != drv.Path)
		{outputStream.WriteLine("Share         = " + drv.Sharename)
		}
		outputStream.WriteLine ("Root path     = " + fileFolderList.rootPath)

		var ftext = []
		if (fileFolderList.filter != ""  ) ftext.push(" " + fileFolderList.filter)
		if (fileFolderList.onlyFolders   ) ftext.push(" no files"                )
		if (fileFolderList.onlyFiles     ) ftext.push(" no folders"              )
		if (fileFolderList.recursive     ) ftext.push(" recursive"               )
		if (fileFolderList.onlyMultiNames) ftext.push(" multiple names only"     )
		outputStream.WriteLine ("Filter        =" + ftext)

		if (fileFolderList.comment != "") outputStream.WriteLine (fileFolderList.comment)

		for (var i = 0; i < emptyTrail; outputStream.WriteLine(""), i++);
	}

// ====================================================================================================================
// Method "showFileFolderStatistics" displays the getFileFolderList statistics.
//
// Arguments:
//	outputStream		An open output TextStream Object, for example WScript.StdOut.
//	fileFolderList		Array containing the result of getFileFolderList.
//	emptyLead		Number of leading empty lines displayed.
//	emptyTrail		Number of trailing empty lines displayed.
//
// Returns:
//	<nothing>
// --------------------------------------------------------------------------------------------------------------------

	hr$dirtool.showFileFolderStatistics = function (outputStream, fileFolderList, emptyLead, emptyTrail)
	{
		function lpad (x, L) {for (var y = "" + x; y.length < L; y  = " " + y); return y}

		var fileCount   = fileFolderList.fileCount
		var folderCount = fileFolderList.folderCount
		var totalSize   = fileFolderList.totalSize
		var driveAvail  = fileFolderList.driveObject.AvailableSpace
		var driveFree   = fileFolderList.driveObject.FreeSpace
		var driveSize   = fileFolderList.driveObject.TotalSize

		var maxNumber  = Math.max (fileCount, folderCount, totalSize, driveAvail, driveFree, driveSize)
		var numWidth   = (""+maxNumber).length

		for (var i = 0; i < emptyLead; outputStream.WriteLine(""), i++);

		outputStream.WriteLine ("File count           = " + lpad(fileCount  ,numWidth))
		outputStream.WriteLine ("Folder count         = " + lpad(folderCount,numWidth))
		outputStream.WriteLine ("Total size           = " + lpad(totalSize  ,numWidth))
		outputStream.WriteLine ("Available free space = " + lpad(driveAvail ,numWidth))
		if (driveFree != driveAvail)
		{	outputStream.WriteLine ("Total free space     = " + lpad(driveFree  ,numWidth))
		}
		outputStream.WriteLine ("Drive size           = " + lpad(driveSize  ,numWidth))

		for (var i = 0; i < emptyTrail; outputStream.WriteLine(""), i++);
	}

// ====================================================================================================================
// Method "showFileFolderInfo" displays the getFileFolderList header, list, and statistics.
//
// Arguments:
//	outputStream		An open output TextStream Object, for example WScript.StdOut.
//	fileFolderList		Array containing the result of getFileFolderList.
//	fieldList		Array with field numbers to be displayed.
//
// Returns:
//	<nothing>
// --------------------------------------------------------------------------------------------------------------------

	hr$dirtool.showFileFolderInfo = function (outputStream,fileFolderList,fieldList,basicOutput)
	{
		if (!basicOutput) hr$dirtool.showFileFolderHeader    (outputStream,fileFolderList,0,1)
		                  hr$dirtool.showFileFolderList      (outputStream,fileFolderList,fieldList,basicOutput)
		if (!basicOutput) hr$dirtool.showFileFolderStatistics(outputStream,fileFolderList,1,0)
	}

// ====================================================================================================================
// Method "dir" does it all:
//	collect and sort the directory information and
//	display a full listing including header and statistics.
//
// Arguments:
//	folderObject		same as for getFileFolderList
//	recursive		same as for getFileFolderList
//	wildcardPattern		same as for getFileFolderList
//	onlyFiles		same as for getFileFolderList
//	onlyFolders		same as for getFileFolderList
//	propertyFilter		same as for getFileFolderList
//	sortItems		same as for sortFileFolderList
//	fieldList		same as for showFileFolderList
//	outputStream		same as for showFileFolderList, -Header, -Statistics
//	comment			any text you like, will be displayed as last line of the heading
//	basicOutput		boolean, if true then only full paths of files and folders are listed, nothing else
//
// Returns:
//	<nothing>
// --------------------------------------------------------------------------------------------------------------------

	hr$dirtool.dir = function
				(folderObject,recursive,wildcardPattern,onlyFiles,onlyFolders,propertyFilter
				,sortItems,fieldList,outputStream,comment,basicOutput
				)
	{
		var startTimeOverall = new Date()

		var fileFolderList = hr$dirtool.getFileFolderList(
				folderObject,recursive,wildcardPattern,onlyFiles,onlyFolders,propertyFilter)

		hr$dirtool.sortFileFolderList (fileFolderList,sortItems)

		if (hr$dirtool.verbose) WScript.StdErr.WriteLine(
			"# Writing result ("+fileFolderList.totalCount+" lines) to output stream...")

		var startTime = new Date()

		fileFolderList.comment = comment
		hr$dirtool.showFileFolderInfo (outputStream,fileFolderList,fieldList,basicOutput)

		if (hr$dirtool.verbose)
		{	var elapsed = (new Date()).valueOf() - startTime.valueOf()
			WScript.StdErr.WriteLine("# Wrote "+fileFolderList.totalCount+" lines to output stream,"
							+ " elapsed: " + (elapsed/1000).toFixed(3) + " sec")
			var elapsed = (new Date()).valueOf() - startTimeOverall.valueOf()
			WScript.StdErr.WriteLine("# Total elapsed: " + (elapsed/1000).toFixed(3) + " sec")
		}

		if (fileFolderList.tempfile != "") FSO.DeleteFile(fileFolderList.tempfile)
	}

// ====================================================================================================================
// Method "dirCommand" interprets Command Line Arguments and then outputs a directory listing.
//
// Function Arguments:
//	getExtraFileFolderInfo		Optional argument.
//					Function to be used for collecting extra file or folder information,
//					as described for the getFileFolderList method.
//
//	extraFileFolderHelpText		Optional argument.
//					A brief description of the extra info produced by getExtraFileFolderInfo.
//					Will be included in help text displayed if "/?" is given on command line.
//
// Returns:
//	<nothing>
//
// Command Line Arguments:
//	- path		the first non-blank argument that does not start with a slash is used as path for which to
//			produce a directory listing. If it is a folder then all content of that folder is displayed,
//			else the file name part is treated as a wildcard pattern for filtering file names.
//
//	- switches	any argument that starts with a slash is considered a switch.
//			see the code below (function showHelp) for information about valid switches.
// --------------------------------------------------------------------------------------------------------------------

	hr$dirtool.dirCommand = function (getExtraFileFolderInfo, extraFileFolderHelpText)
	{
		if (arguments.length > 0) hr$dirtool.getExtraFileFolderInfo = getExtraFileFolderInfo
		var extraHelp =
			( arguments.length > 1
			? extraFileFolderHelpText
			: "Show extra file/folder information (if implemented)"
			)

		var path		= ""
		hr$dirtool.useExtraInfo = false
		var exclude		= false
		var recursive		= false
		var onlyFiles		= false
		var onlyFolders		= false
		var sortItems		= []
		var fieldList		=
					[hr$dirtool.FileNameField
					,hr$dirtool.SizeField
					,hr$dirtool.ExtraInfoField
					,hr$dirtool.DateLastModifiedField
					,hr$dirtool.AttributesField
					,hr$dirtool.FolderNameField
					]
		var outputStream	= WScript.StdOut
		var comment		= ""
		var basicOutput		= false
		hr$dirtool.verbose	= false
		var onlyMultiNames	= false
		var onlyMultiExtra	= false

		var inDate	= ""
		var exDate	= ""
		var loDate	= ""
		var hiDate	= ""

		var minsize = Number.NEGATIVE_INFINITY
		var maxsize = Number.POSITIVE_INFINITY

		var attrORincl   = NaN
		var attrORexcl   = NaN
		var attrANDincl  = NaN
		var attrANDexcl  = NaN

		function showHelp()
		{
			outputStream.WriteLine("hr$dirtool V" + hr$dirtool.version)
			outputStream.WriteLine (""
			+ "\n  Script: " + WScript.ScriptFullName
			+ "\n"
			+ "\n  Usage:"
			+ "\n    Cscript " + WScript.ScriptName
				+ " {pattern | folderpath | folderpath\\pattern} [/switch ...]"
			+ "\n"
			+ "\n  Pattern:"
			+ "\n    wildcard pattern for filtering names, valid wildcards are '*' and '?'"
			+ "\n"
			+ "\n  General switches:"
			+ "\n    /A=xx     List only files with certain attributes set"
			+ "\n    /-A=xx    Exclude files with certain attributes set"
			+ "\n              'xx' is a string of attribute characters (rhsdalc) on which to filter."
			+ "\n              by default a boolen OR is applied, e.g. /A=HS lists files with either"
			+ "\n              H or S or both set. if 'xx' is preceded by a plus sign then a boolean"
			+ "\n              AND is applied, e.g. /A=+HS lists files with both H and S set."
			+ "\n              You can specify all combinations of /A=xx /-A=xx /A=+xx and /-A=+xx,"
			+ "\n              only files passing all filters are listed."
			+ "\n"
			+ "\n    /B        Basic output, only full path names are listed"
			+ "\n"
			+ "\n    /F        List only files, no folders"
			+ "\n    /-F       List only folders, no files"
			+ "\n              if /-F is used without /S then folder sizes are displayed"
			+ "\n"
			+ "\n    /FFF=filename  specify a 'File/Folder Filter File'"
			+ "\n              The file may contain the following sections:"
			+ "\n              [arguments]       each line is interpreted as a normal switch"
			+ "\n              [includefiles]    each line is a Regular Expression to include files"
			+ "\n              [excludefiles]    each line is a Regular Expression to exclude files"
			+ "\n              [includefolders]  each line is a Regular Expression to include folders"
			+ "\n              [excludefolders]  each line is a Regular Expression to exclude folders"
			+ "\n              for more info about Regular Expressions you can for example visit page"
			+ "\n                2380d458-3366-402b-996c-9363906a7353.asp"
			+ "\n                at http://msdn.microsoft.com/library/en-us/script56/html/"
			+ "\n              (Regular Expressions are VERY powerful, although practically unreadable)."
			+ "\n              Please note that these filters are applied to the FULL PATH of a file"
			+ "\n              or folder and that folders excluded by one of these filters are not"
			+ "\n              recursed into, not even if the /S switch is used."
			+ "\n"
			+ "\n    /M        List only files or folders with a name that occurs more than once."
			+ "\n              '/M' implies '/S'"
			+ "\n"
			+ "\n    /M+       List only files or folders with extra info (i.e. the info listed when"
			+ "\n              '/+' is used) that occurs more than once. '/M'+ implies '/+'."
			+ "\n              If, for example, the extra info is a file's MD5 hash value, then you will"
			+ "\n              be able to find all different files with identical content!  But please"
			+ "\n              be aware that in that case ALL FILES will be read to determine the MD5,"
			+ "\n              so when scanning an entire disk it will need some time..."
			+ "\n"
			+ "\n    /O=items  Sort by given keys in order of appearance:"
			+ "\n              A(rhsdalc)=attributes, D=date, E=extension, F=folder, N=name, S=size"
								+ (arguments.length > 0 ? ", X=extraInfo" : "")
			+ '\n              prefix with "-" for descending sort'
			+ "\n"
			+ "\n    /S        Recurse into subfolders"
			+ "\n"
			+ "\n    /T        Do not use a temporary file for storage of intermediate data."
			+ "\n              Default behaviour is using a temporary file, which is recommended"
			+ "\n              especially if a large amount of files is expected to be listed;"
			+ "\n              not using a temporary file dramatically increases memory usage"
			+ "\n              and causes an unacceptable slow down of overall performance."
			+ "\n              See http://henk-reints.nl/jscript/htm/jsarrayperf.htm"
			+ "\n"
			+ "\n    /T=path   Use 'path' as temporary file instead of the default"
			+ "\n              (which is " + hr$dirtool.tempfile + ")"
			+ "\n              Please note that some extra files may be temporarily created, using the"
			+ "\n              same name suffixed with something extra like a sequence number or so."
			+ "\n"
			+ "\n    /U=nnn    Set sort buffer size, only applicable if a temporary file is used."
			+ "\n              Current default = " + hr$dirtool.sortBufferSize
			+ "\n              A small value will cause many intermediate temporary files to be created"
			+ "\n              during sorting, but a large buffer may significantly slow down the sorting"
			+ "\n              process since JavaScript Arrays are very inefficient."
			+ "\n"
			+ "\n    /V        Verbose, i.e. write progress info to StdErr"
			+ "\n"
			+ "\n    /X        Use pattern as an exclusion filter (list only file names that DON'T match)"
			+ "\n"
			+ "\n    /minsize=bytes  List only if greater than or equal to size"
			+ "\n    /maxsize=bytes  List only if smaller than or equal to size"
			+ "\n                    if   minsize <= maxsize"
			+ "\n                    then filter = 'filesize >= minsize AND filesize <= maxsize'"
			+ "\n                    else filter = 'filesize >= minsize OR  filesize <= maxsize'"
			+ "\n"
			+ "\n    /+        " + extraHelp
			+ "\n"
			+ "\n    /?        Display this help text"
			+ "\n"
			+ "\n  Date filtering switches:"
			+ "\n    /date=datestring    List only if filedate matches datestring"
			+ "\n    /-date=datestring   List only if filedate does not match datestring"
			+ "\n    /since=datestring   List only if filedate greater than or equal to datestring"
			+ "\n    /before=datestring  List only if filedate less than datestring"
			+ "\n    use '/?D' for more detailed help"
			)
			if (Date.applyFileDateCorrection) outputStream.WriteLine (""
			+ "\n  Note:"
			+ "\n    The file dates are corrected for the Windows flaw of applying the CURRENT system"
			+ "\n    timezone to every file or folder date, even if it is in another DST period."
			)
			outputStream.WriteLine (""
			+ "\n  Tip: Make your command prompt window as large as possible:"
			+ "\n       Right-click on the Title bar => Properties => Layout"
			)
			WScript.Quit()
		}

		function showDateHelp()
		{
			function parseDate(dateString)
			{	var ds = hr$dirtool.parseDate(dateString), L = ds.length
				for (; ds.length < 15; ds+=" ");
				var dY = 1*ds.substr( 0,4); if (!isFinite(dY) || L <  4) dY = 1970
				var dM = 1*ds.substr( 4,2); if (!isFinite(dM) || L <  6) dM = 1
				var dD = 1*ds.substr( 6,2); if (!isFinite(dD) || L <  8) dD = 1
				var dH = 1*ds.substr( 8,2); if (!isFinite(dH) || L < 10) dH = 0
				var dN = 1*ds.substr(10,2); if (!isFinite(dN) || L < 12) dN = 0
				var dS = 1*ds.substr(12,2); if (!isFinite(dS) || L < 14) dS = 0
				var mo = "Jan/Feb/Mar/Apr/May/Jun/Jul/Aug/Sep/Oct/Nov/Dec".split("/")
				var Y = ("000" + dY).slice(-4)
				var M = mo[dM-1] + "-"
				var D = ("0"+dD).slice(-2) + "-"
				var H = " " + ("0"+dH).slice(-2)
				var N = ":" + ("0"+dN).slice(-2)
				var S = ":" + ("0"+dS).slice(-2)
				if (L ==  4) ds += "(year only)"
				if (L ==  6) ds += "= " + M+Y
				if (L ==  8) ds += "= " + D+M+Y
				if (L == 10) ds += "= " + D+M+Y+H+"o'clock"
				if (L == 12) ds += "= " + D+M+Y+H+N
				if (L == 14) ds += "= " + D+M+Y+H+N+S
				return ds
			}
			outputStream.WriteLine("hr$dirtool V" + hr$dirtool.version)
			outputStream.WriteLine (""
			+ "\n  Date filter options:"
			+ "\n"
			+ "\n    /date=datestring    List only if filedate matches datestring"
			+ "\n    /-date=datestring   List only if filedate does not match datestring"
			+ "\n    /since=datestring   List only if filedate greater than or equal to datestring"
			+ "\n    /before=datestring  List only if filedate less than datestring"
			+ "\n"
			+ "\n  if   sincedate < beforedate"
			+ "\n  then filter = 'filedate >= sincedate AND filedate < beforedate'"
			+ "\n  else filter = 'filedate >= sincedate OR  filedate < beforedate'"
			+ "\n"
			+ "\n  datestring is expected to be of form  yyyymmddhhnnss  (hh=24hoursystem, nn=minutes)"
			+ "\n  it can be truncated to a shorter string for reduced accuracy, for example a date only"
			+ "\n"
			+ "\n  Special datestring components:"
			+ "\n"
			+ "\n    substrings:"
			+ "\n      'today', 'tomorrow', 'yesterday', 'now', 'time'"
			+ "\n      'lastsun'..'lastsat', 'nextsun'..'nextsat'"
			+ "\n      'lastjan'..'lastdec', 'nextjan'..'nextdec'"
			+ "\n    are replaced with their respective date and/or time values,"
			+ "\n    'lastxxx' and 'nextxxx' are relative to the current system date"
			+ "\n    'now' = full current date/time, 'time' is current time w/o date"
			+ "\n"
			+ "\n    substrings: 'yyyy' 'mm' 'dd' 'hh' 'nn' 'ss'  (nn=minutes)"
			+ "\n    are replaced with the corresponding component from the current system date/time"
			+ "\n"
			+ "\n    substrings of the form '+nnU' or '-nnU'"
			+ "\n    where 'nn' is a number and 'U' is one of the units Y M W D H N S"
			+ "\n    are relative components and will add or subtract the given number of units"
			+ "\n"
			+ "\n    after substitution of special substrings all non-digits are removed,"
			+ "\n    so you can freely use non-interpreted characters for interpunction"
			+ "\n"
			+ "\n    PLEASE NOTE:"
			+ "\n    only the VALUE is substituted in the datestring, but its POSITION determines"
			+ "\n    its meaning, i.e. which date component it is, e.g. specifying only 'time' will"
			+ "\n    will result in 'hhnnss' being interpreted as MONTH 'ss' of the YEAR 'hhnn'!"
			+ "\n"
			+ "\n  Final accuracy is conform smallest unit used in absolute or relative components"
			+ "\n"
			+ "\n  Examples:"
			+ "\n"
			)
			var examples =
				["now          "
				,"now-1h       "
				,"today        "
				,"todaytime    "
				,"today-3d     "
				,"tomorrow+1w  "
				,"yesterday1430"
				,"lastmon-1m   "
				,"yyyymm       "
				,"lastnov18    "
				,"nextdec_25   "
				,"nextwed_time "
				,"time_today   "	// time_today must be the last example!
				]
			for (var i = 0; i < examples.length; i++)
			{	outputStream.WriteLine(
					"    /D=" + examples[i] + "  => " + parseDate(examples[i])
					)
			}
			outputStream.WriteLine(""
			+ "\n      (please take good notice of the last example!"
		//	+ " it may seem to incorrectly handle"
		//	+ "\n      the time at, say, 09:50:45, but that will be month 45 of year 950, which of course"
		//	+ "\n      actually is month 9 of the year 953. You don't have files from that date, do you?)"
			)
			WScript.Quit()
		}

		function fatal (msg)
		{
			WScript.StdErr.WriteLine ("# Error: " + msg)
			WScript.StdErr.WriteLine ('# (use "/?" for help)')
			WScript.Quit(1)
		}

		function parseAttrMask(attrString)
		{	var m = 0
			var x = attrString.toLowerCase()
			for (var i = 0; i < x.length; i++)
			{	var k = hr$dirtool.attrCharacters.indexOf(x.charAt(i))
				if (k >= 0) m |= (1<<k)
			}
			return m
		}

		function parseAttrSort(attrSortString)
		{	var attrSort = []
			for (var i = 0; i < attrSortString.length; i++)
			{	var c = attrSortString.charAt(i).toUpperCase(), f = 0, i1 = i-1, ok = true
				switch (c)
				{	case "R": f = hr$dirtool.attrReadOnly;		break
					case "H": f = hr$dirtool.attrHidden;		break
					case "S": f = hr$dirtool.attrSystem;		break
					case "V": f = hr$dirtool.attrVolumeID;		break
					case "D": f = hr$dirtool.attrDirectory;		break
					case "A": f = hr$dirtool.attrArchive;		break
					case "L": f = hr$dirtool.attrLink;		break
					case "C": f = hr$dirtool.attrCompressed;	break
					case "-": ok = (i < 1 || attrSortString.charAt(i1) != "-");	break
					default:  ok = false;						break
				}
				if (!ok) fatal ("Invalid sort attribute: " + c)
				if (f > 0)
				{	if (i > 0 && attrSortString.charAt(i1) == "-") f = -f
					attrSort.push(f)
				}
			}
			return attrSort
		}

		function parseSortItems(sortItemString)
		{	var sortItems = []
			for (var i = 0; i < sortItemString.length; i++)
			{	var c = sortItemString.charAt(i).toUpperCase(), f = 0, i1 = i-1, ok = true
				switch (c)
				{	case "A":
						var j = i + 1, k = -1
						if (sortItemString.charAt(j)=="(") k = sortItemString.indexOf(")",++j)
						if (k > 0)
						{	i = k
							f = hr$dirtool.AttributesField
							sortItems.attrSort = parseAttrSort(sortItemString.slice(j,k))
						} else	fatal ("Invalid sort attributes string: "
								+ sortItemString.slice(i) )
						break
					case "D": f = hr$dirtool.DateLastModifiedField;	break
					case "E": f = hr$dirtool.FileTypeField;		break
					case "F": f = hr$dirtool.FolderNameField;	break
					case "N": f = hr$dirtool.FileNameField;		break
					case "S": f = hr$dirtool.SizeField;		break
					case "X": f = hr$dirtool.ExtraInfoField;	break
					case "-": ok = (i < 1 || sortItemString.charAt(i1) != "-");	break
					default:  ok = false;						break
				}
				if (!ok) fatal ("Invalid sort field: " + c)
				if (f > 0)
				{	if (i > 0 && sortItemString.charAt(i1) == "-") f = -f
					sortItems.push(f)
				}
			}
			return sortItems
		}

		var argList        = []
		var includeFiles   = []
		var excludeFiles   = []
		var includeFolders = []
		var excludeFolders = []

		function fffRead (fileName)
		{	var e
			if (!FSO.FileExists(fileName)) fatal ("File not found: /fff=" + fileName)
			try {var fffStream = FSO.OpenTextFile(fileName,1) }
			catch(e) {fatal("Unable to open file /fff=" + fileName + " [" + e.description + "]") }
			var section = ""
			while (!fffStream.AtEndOfStream)
			{	try {var fffLine = fffStream.ReadLine() }
				catch(e) {fatal("Unable to read file /fff=" + fffName + " [" + e.description + "]") }
				fffLine = fffLine.replace(/^\s+|\s+$/g,"")
				if (fffLine.charAt(0) == "[" && fffLine.slice(-1) == "]")
					section = fffLine.toLowerCase()
				else if (fffLine != "") switch (section)
				{	case "[arguments]":		putarg(fffLine);		break
					case "[includefiles]":		includeFiles.push(fffLine);	break
					case "[excludefiles]":		excludeFiles.push(fffLine);	break
					case "[includefolders]":	includeFolders.push(fffLine);	break
					case "[excludefolders]":	excludeFolders.push(fffLine);	break
				}
			}
			fffStream.Close()
		}

		function putarg (arg)
		{	if (arg.toUpperCase().indexOf("/FFF=") != 0)
				argList.push(arg)
			else	fffRead(arg.slice(5))
		}

	// collect all arguments. if we have been invoked by 'ddir.hta' then the command
	// line arguments are in Array 'WScript.HTAParameter', so try that first:
		var e
		try	{for (var i = 0; i < WScript.HTAParameter.length; putarg(WScript.HTAParameter[i++]) );}
		catch(e){for (var i = 0; i < WScript.Arguments.length; putarg(WScript.Arguments(i++)) );}

	// parse the collected arguments:
		for (var i = 0; i < argList.length; i++)
		{	var arg = argList[i]
			if (arg.charAt(0) != "/")
			{	if (path == "") path = arg
			} else
			{	var argSplit = arg.slice(1).split("=")
				var argName  = argSplit.shift().toUpperCase()
				var argValue = argSplit.join("=")

			// evt. nog checken of value ontbreekt of ten onrechte en indien ja dan fatal
				switch (argName)
				{
					case "A":	if (argValue.charAt(0) == "+")
								attrANDincl = parseAttrMask(argValue)
							else	attrORincl  = parseAttrMask(argValue)
							break

					case "-A":	if (argValue.charAt(0) == "+")
								attrANDexcl = parseAttrMask(argValue)
							else	attrORexcl  = parseAttrMask(argValue)
							break

					case "B":	basicOutput = true;				break

					case "DATE":	inDate = hr$dirtool.parseDate(argValue);	break
					case "-DATE":	exDate = hr$dirtool.parseDate(argValue);	break
					case "SINCE":	loDate = hr$dirtool.parseDate(argValue);	break
					case "BEFORE":	hiDate = hr$dirtool.parseDate(argValue);	break

					case "F":	onlyFiles = true;				break
					case "-F":	onlyFolders = true;				break

					case "M":	onlyMultiNames = true
							recursive = true
							break

					case "M+":	onlyMultiExtra = true
							hr$dirtool.useExtraInfo = true
							break

					case "O":	sortItems = parseSortItems(argValue);		break
					case "S":	recursive = true;				break

					case "T":	if (argValue == "")
								hr$dirtool.tempfile = ""
							else
							{	var tempfile = FSO.GetAbsolutePathName(argValue)
								if (FSO.FolderExists(tempfile))
								  tempfile+="\\"+FSO.GetFileName(hr$dirtool.tempfile);
								hr$dirtool.tempfile = tempfile
							}
							break

					case "U":	var u = parseInt(argValue,10)
							if (isFinite(u) && u > 1) hr$dirtool.sortBufferSize = u
							break

					case "V":	hr$dirtool.verbose = true;		break
					case "X":	exclude = true;				break
					case "MINSIZE":	minsize = parseInt(argValue);		break
					case "MAXSIZE":	maxsize = parseInt(argValue);		break
					case "+":	hr$dirtool.useExtraInfo = true;		break

					case "?D":	showDateHelp();				break
					case "?":	showHelp();				break

					default:	fatal ("Invalid switch: " + argName);	break
				}
			}
		}

		if (onlyMultiNames && onlyMultiExtra) fatal ("Conflicting switches: /M /M+")

	// set comment if certain switches used:
		// (RE patterns are also translated to true Regular Expressions)
		var REfilters =
			[["in","file"  ,includeFiles  ]
			,["ex","file"  ,excludeFiles  ]
			,["in","folder",includeFolders]
			,["ex","folder",excludeFolders]
			]
		for (var j = 0; j < REfilters.length; j++)
		{	if (REfilters[j][2].length > 0)
			{	if (comment != "") comment += "\n"
				comment += "  "+REfilters[j][0]+"clusive RE "+REfilters[j][1]+" filter(s):"
				for (var i = 0; i < REfilters[j][2].length; i++)
				{	comment += "\n    " + REfilters[j][2][i]
					REfilters[j][2][i] = new RegExp(REfilters[j][2][i],"i")
				}
			}
		}

		if (inDate+exDate+loDate+hiDate != "")
		{	if (comment != "") comment += "\n"
			comment += "Date filter   ="
			if (inDate != "") comment += " /date=" + inDate
			if (exDate != "") comment += " /-date=" + exDate
			if (loDate != "") comment += " /since=" + loDate
			if (hiDate != "") comment += " /before=" + hiDate
		}

		if (isNaN(minsize)) minsize = Number.NEGATIVE_INFINITY
		if (isNaN(maxsize)) maxsize = Number.POSITIVE_INFINITY
		if (minsize > Number.NEGATIVE_INFINITY || maxsize < Number.POSITIVE_INFINITY)
		{	if (comment != "") comment += "\n"
			comment += "Size filter   ="
			if (minsize > Number.NEGATIVE_INFINITY) comment += " /minsize="+minsize
			if (maxsize < Number.POSITIVE_INFINITY) comment += " /maxsize="+maxsize
		}

		if (!isNaN(attrORincl) || !isNaN(attrORexcl) || !isNaN(attrANDincl) || !isNaN(attrANDexcl) )
		{	if (comment != "") comment += "\n"
			comment += "Attr filter   ="
			if (!isNaN(attrORincl )) comment += " /A="   + hr$dirtool.fmtAttr(attrORincl ).replace(/-/g,"")
			if (!isNaN(attrORexcl )) comment += " /-A="  + hr$dirtool.fmtAttr(attrORexcl ).replace(/-/g,"")
			if (!isNaN(attrANDincl)) comment += " /A=+"  + hr$dirtool.fmtAttr(attrANDincl).replace(/-/g,"")
			if (!isNaN(attrANDexcl)) comment += " /-A=+" + hr$dirtool.fmtAttr(attrANDexcl).replace(/-/g,"")
		}

	// define the file and folder filter functions:
		function fileFilter (fileObject)
		{	var i, incl = (includeFiles.length == 0), excl = false, path = fileObject.Path
			for (i=0; !incl && i < includeFiles.length; incl=includeFiles[i++].test(path) );
			for (i=0; !excl && i < excludeFiles.length; excl=excludeFiles[i++].test(path) );
			return (incl && !excl)
		}
		function folderFilter (folderObject)
		{	var i, incl = (includeFolders.length == 0), excl = false, path = folderObject.Path
			for (i=0; !incl && i < includeFolders.length; incl=includeFolders[i++].test(path) );
			for (i=0; !excl && i < excludeFolders.length; excl=excludeFolders[i++].test(path) );
			return (incl && !excl)
		}
		hr$dirtool.fileFilter   = fileFilter
		hr$dirtool.folderFilter = folderFilter

	// preprocess the given path:
		if (path.slice(-1) == ":") path += "\\"
		var mustBeFolder = path.slice(-1) == "\\"

		if (path == "") path = "."
		var path = FSO.GetAbsolutePathName(path)

		if (FSO.FolderExists(path))
		{	var folderObject    = FSO.GetFolder(path)
			var wildcardPattern = ""
		} else if (mustBeFolder)
		{	fatal ("Folder does not exist: " + path)
		} else
		{	var folderPath = FSO.GetParentFolderName(path)
			var wildcardPattern = (exclude ? "/" : "") + FSO.GetFileName(path)
			if (FSO.FolderExists(folderPath))
				var folderObject = FSO.GetFolder(folderPath)
			else	fatal ("Folder does not exist: " + folderPath)
		}

		if (hr$dirtool.verbose) WScript.StdErr.WriteLine ("# hr$dirtool V" + hr$dirtool.version)

	// go list the directory:
		hr$dirtool.dir	(folderObject,recursive,wildcardPattern,onlyFiles,onlyFolders,
				[inDate,exDate,loDate,hiDate
				,minsize,maxsize
				,attrORincl,attrORexcl,attrANDincl,attrANDexcl
				,onlyMultiNames,onlyMultiExtra
				]
				,sortItems,fieldList,outputStream,comment,basicOutput)
	}

// ====================================================================================================================
// Method "createFolder" creates a new folder, including the entire path to the new folder if necessary.
//
// Arguments:
//	folderPath		String containing the relative or absolute path of the folder to be created.
//
//	verbose			Boolean value indicating whether to display informational messages.
//				Error messages are always written and cannot be suppressed.
//
//	internal		Boolean for recursion. You should not use this at the user level invocation.
//
// Returns:
//	<nothing>
//
// Notes:
//	This function assumes the GLOBAL variable FSO has been defined as a FileSystemObject Object.
//	It writes messages to the WScript.StdErr Stream, so it works with Cscript only.
// --------------------------------------------------------------------------------------------------------------------

	hr$dirtool.createFolder = function (folderPath, verbose, internal)
	{
		var quiet = (internal ? true : false)
		var dolog = (verbose  ? true : false)
		if (dolog && !quiet) WScript.StdErr.WriteLine('# Preparing to create folder "' + folderPath + '"')
		var parentName = FSO.GetParentFolderName(folderPath)
		if (parentName != "")
		{	if (!FSO.FolderExists(parentName))
			{	if (parentName.slice(-1) == "\\")
				{	WScript.StdErr.WriteLine('#  root folder "' + parentName + '" does not exist')
					dolog = true
				}
				hr$dirtool.createFolder (parentName, dolog, true)
			}
			if (!FSO.FolderExists(folderPath))
			{	if (dolog) WScript.StdErr.WriteLine('#  creating folder "' + folderPath + '"')
				FSO.CreateFolder(folderPath)
			} else
			{	if (dolog) WScript.StdErr.WriteLine('#  folder "' + folderPath + '" already exists')
			}
		}
	}

// ====================================================================================================================
// some methods intended for internal use, but of course you're free to use them to your advantage:
// --------------------------------------------------------------------------------------------------------------------

	hr$dirtool.fmtAttr = function (attr)					// formatted attributes string
	{	var result = "", chars = hr$dirtool.attrCharacters
		for (var m=1, i=0; i < chars.length; m<<=1, i++)
		{	if (m != hr$dirtool.attrVolumeID) result += (attr&m ? chars.charAt(i) : "-")
		}
		return result
	}

	hr$dirtool.fmtDate = function (dateValue)				// date string YYYY-MM-DD HH:mm:ss
	{	var d = new Date(dateValue)
		var Y = (""+(d.getFullYear()+10000)).slice(-4)	// assume a positive 4-digit year
		var M = (""+(d.getMonth()   +  101)).slice(-2)	// JS month number is 0..11
		var D = (""+(d.getDate()    +  100)).slice(-2)
		var H = (""+(d.getHours()   +  100)).slice(-2)
		var m = (""+(d.getMinutes() +  100)).slice(-2)
		var s = (""+(d.getSeconds() +  100)).slice(-2)
		return [[Y,M,D].join("-"),[H,m,s].join(":")].join(" ")
	}

	hr$dirtool.fmtDate0 = function (dateValue)				// date string YYYYMMDDHHmmss
	{	return hr$dirtool.fmtDate(dateValue).replace(/[^0-9]/g,"")
	}

// ====================================================================================================================
// below is the date string parser used to interpret date filter values:
// --------------------------------------------------------------------------------------------------------------------

	hr$dirtool.parseDate = function (dateString)				// parse cmd line date arg
	{
		function fmt(y,m,d) {return hr$dirtool.fmtDate0((new Date(y,m,d)).valueOf()).slice(0,8)}

	// define special dates and date components (today/tomorrow/yesterday, lastsun-sat, nextsun-sat):
		var now = new Date()
		var Y = now.getFullYear(), M = now.getMonth()  , D = now.getDate()   , W = now.getDay()
		var H = now.getHours()   , N = now.getMinutes(), S = now.getSeconds()
		var nowstring  = hr$dirtool.fmtDate0(now.valueOf())
		var timestring = nowstring.slice(8)
		var data =
			[[fmt(Y,M,D  )		,"today"]
			,[fmt(Y,M,D+1)		,"tomorrow"]
			,[fmt(Y,M,D-1)		,"yesterday"]
			,[("000"+Y).slice(-4)	,"yyyy"]
			,[("0"+(M+1)).slice(-2)	,"mm"]
			,[("0"+D).slice(-2)	,"dd"]
			,[("0"+H).slice(-2)	,"hh"]
			,[("0"+N).slice(-2)	,"nn"]
			,[("0"+S).slice(-2)	,"ss"]
			,[nowstring		,"now"]
			,[timestring		,"time"]
			]
		var dd, days = "sun/mon/tue/wed/thu/fri/sat".split("/")
		for (var i = 0; i < days.length; i++)
		{	dd = (W + 6 - i) % 7 + 1; data.push( [fmt(Y,M,D-dd), "last"+days[i] ] );
			dd = (i + 6 - W) % 7 + 1; data.push( [fmt(Y,M,D+dd), "next"+days[i] ] );
		}
		var mm, months = "jan/feb/mar/apr/may/jun/jul/aug/sep/oct/nov/dec".split("/")
		for (var i = 0; i < months.length; i++)
		{	mm = (M + 11 - i) % 12 + 1; data.push( [fmt(Y,M-mm,1).slice(0,6), "last"+months[i] ] );
			mm = (i + 11 - M) % 12 + 1; data.push( [fmt(Y,M+mm,1).slice(0,6), "next"+months[i] ] );
		}
		//data.sort()	// sorting is meaningful for debugging only
		var j,k, ds = dateString.toLowerCase()
		if (ds == "") ds = "today"

	// collect all relative offsets within the resulting date string and remove them thereof:
		var ofs = []
		for (var i = 0; i < ds.length; )
		{	var k = -1
			if ("+-".indexOf(ds.charAt(i)) >= 0)				// "+" or "-"
			{	for (var j = i+1; k < 0 && j < ds.length; j++)
				{ if ("ymwdhns".indexOf(ds.charAt(j)) >= 0) k = j+1	// up to unit character
				}
			}
			if (k < 0) i++
			else
			{	ofs.push(ds.slice(i,k))			// save offset string
				ds = ds.slice(0,i) + ds.slice(k)	// and remove it
			}
		}

	// replace special dates or date components (defined in round 0) with appropriate values:
		for (var i = 0; i < data.length; i++)
		{	while (true)
			{	j = ds.indexOf(data[i][1])
				if (j < 0) break
				k = data[i][1].length
				ds = ds.slice(0,j) + data[i][0] + ds.slice(j+k)	
			}
		}
		ds = ds.replace(/[^0-9]/g,"")						// remove all non-digits
		var L = ds.length; L -= L%2; if (L < 4) L = 0; ds = ds.slice(0,L)	// remove incomplete fields

	// apply collected relative offsets:
		var dY = 1*ds.substr( 0,4); if (!isFinite(dY) || L <  4) dY = Y
		var dM = 1*ds.substr( 4,2); if (!isFinite(dM) || L <  6) dM = M+1
		var dD = 1*ds.substr( 6,2); if (!isFinite(dD) || L <  8) dD = D
		var dH = 1*ds.substr( 8,2); if (!isFinite(dH) || L < 10) dH = 0
		var dN = 1*ds.substr(10,2); if (!isFinite(dN) || L < 12) dN = 0
		var dS = 1*ds.substr(12,2); if (!isFinite(dS) || L < 14) dS = 0
		for (var i = 0; i < ofs.length; i++)
		{	var delta = parseInt(ofs[i],10), unit = ofs[i].slice(-1)
			if (isFinite(delta))
			{	delta = Math.round(delta)
				if (unit == "y") {dY += delta  ; L = Math.max(L, 4) }
				if (unit == "m") {dM += delta  ; L = Math.max(L, 6) }
				if (unit == "w") {dD += delta*7; L = Math.max(L, 8) }
				if (unit == "d") {dD += delta  ; L = Math.max(L, 8) }
				if (unit == "h") {dH += delta  ; L = Math.max(L,10) }
				if (unit == "n") {dN += delta  ; L = Math.max(L,12) }
				if (unit == "s") {dS += delta  ; L = Math.max(L,14) }
			}
		}
		if (L < 14) dS = 0
		if (L < 12) dN = 0
		if (L < 10) dH = 0
		if (L <  8) dD = 1
		if (L <  6) dM = 1
		if (L <  4) dY = 1970
		ds = hr$dirtool.fmtDate0( (new Date(dY,dM-1,dD,dH,dN,dS)).valueOf() ).slice(0,L)
		return ds
	}

// ====================================================================================================================
// finally we define the default tempfile path:
// --------------------------------------------------------------------------------------------------------------------

	hr$dirtool.tempfile = FSO.BuildPath(FSO.GetSpecialFolder(2).Path
				,"hr$dirtool"+hr$dirtool.fmtDate0((new Date()).valueOf())+".tmp")

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// END-OF-FILE  --  hr$dirtool.js  --  Copyright (c) 2005 Henk Reints, http://henk-reints.nl
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
