///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// File:	hr$wsh-fso.js
//
// Purpose:	Define a WshShell Object and a FileSystemObject Object.
//
// Date:	November 2005.
// Last update:	06-JUL-2007.
//
// Author:	Copyright (c) 2005,2007 Henk Reints, http://henk-reints.nl
//
// Prerequisites:
//		none.
//
// DISCLAIMER:
//	HENK REINTS HAS WRITTEN THIS SOFWARE FOR HIS OWN USE, AS A HOBBY, WITHOUT ANY COMMERCIAL INTENTIONS.
//	IT IS PROVIDED AS IS, WITHOUT ANY GUARANTEE FOR CORRECTNESS OR USABILITY. USAGE IS AT YOUR OWN RISK.
//	IN NO WAY SHALL HENK REINTS BE LIABLE FOR ANY NEGATIVE CONSEQUENCES OF USING THIS SOFTWARE.
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	var WSH = new ActiveXObject ("WScript.Shell")
	var FSO = new ActiveXObject ("Scripting.FileSystemObject")

// -------------------------------------------------------------------------------------------------
// Update 05-JUL-2007: Created method Date.parseFileDate.
// This method corrects the following flaw in Windows: file dates are stored as UTC in the
// directory on disk, but when retrieving a file date Windows adds the CURRENT timezone of the
// system time to it, producing a one-hour error if the file date and system time are in a different
// DST period. The Date.parseFileDate method parses the file date property, replaces the timezone
// with the current system timezone and parses it once again, thus returning the true file date
// value.
//
// Although the chance is very very very low, it will however return a wrong file date if a DST
// switch occurs just inbetween, i.e. if Windows applies the before-DST-switch system timezone to
// the file date and Date.parseFileDate uses the after-DST-switch system timezone.
// If you are worried about this to occur it can be prevented as follows:
//
//	do
//	{	var tz = (new Date()).getTimezoneOffset()
//		var fileObj = FSO.GetFile("myfile.dat")
//		var fileDate = Date.parseFileDate(fileObj.DateLastModified)
//	} while ((new Date()).getTimezoneOffset() != tz)
//
// Input should be one of the DateCreated, DateLastAccessed, or DateLastModified Properties of
// a File or Folder Object as returned by FSO.GetFile, FSO.GetFolder, or other FSO methods.
// These properties pretend to be a date (typeof returns "date") but they are definitely NOT
// JavaScript Date Objects. Also in VBS they do not behave as a true date type.
//
// The returned value is the result of Date.parse, so a Number with an internal Date Object value.
// With the boolean property 'Date.applyFileDateCorrection' the correction can be turned on or off.
// -------------------------------------------------------------------------------------------------

	Date.applyFileDateCorrection = true

	Date.parseFileDate = function (fileDateProperty)
	{	var x = Date.parse(fileDateProperty)		// parse the input
		if (!isFinite(x) || !Date.applyFileDateCorrection) return x	// get out if no correction to do
		var f = (new Date(x)).toString().split(" ")	// convert JS Date and split into pieces
		var t = (new Date( )).toString().split(" ")	// split current system time into pieces
		f[4] = t[4]					// copy current system timezone to file date
		return Date.parse(f.join(" "))			// rebuild file date string and parse once again
	}

// --------------------------------------------------------------------------------------------------------
// Update 06-JUL-2007: Created method Date.checkFileDateFlaw.
// It might be possible that on some systems or in future Windows versions the file date flaw will
// be corrected. The method below can be invoked to check for the flaw.
// It uses the 'touch.exe' tool from GNU for Windows in order to create a file with a data in the
// other DST period as the current system time. Then it retrieves the file date and checks if
// the time-of-day is in error or not and sets the Date.applyFileDateCorrection accordingly.
//
// input argument must be the path of the folder where touch.exe resides.
// if it cannot find touch.exe then it returns without doing anything.
// if no input is given then touch.exe is searched in the subfolder tree of where the currently running
// script resides.
//
// Date.checkFileDateFlaw(touchLocation) needs only a single call at the very start of your calling script.
// --------------------------------------------------------------------------------------------------------

	Date.checkFileDateFlaw = function (touchLocation)
	{	var touch = "touch.exe"
		if (arguments.length > 0)
		{	touch = FSO.BuildPath (touchLocation, touch)
			if (!FSO.FileExists(touch)) return
		} else
		{	var here = FSO.GetParentFolderName(WScript.ScriptFullName)
			touch = FSO.BuildPath (here, touch)
			var command = "cmd /c dir /s /b " + touch
			var shell = WSH.Exec(command)
			for (touch = ""; touch == "" && !shell.StdOut.AtEndOfStream; touch = shell.StdOut.ReadLine() );
			if (touch == "") return
		}
		var now  = new Date()
		var year = now.getFullYear()
		var hour = 12
		var tz0  = now.getTimezoneOffset()*-1
		var tz1  = (new Date(year,0,1,hour,0,0)).getTimezoneOffset()*-1	// timezone on Jan 1
		var tz2  = (new Date(year,6,1,hour,0,0)).getTimezoneOffset()*-1	// timezone on Jul 1
		if (tz1 == tz2)					// if both are equal then
		{	Date.applyFileDateCorrection = false	// DST is not used on this computer
			return
		}
		if (Math.max(tz1,tz2) == tz0)
		{	var month = (tz2 > tz1 ? "01" : "07")	// North/South winter month if DST currently on
		} else	var month = (tz2 > tz1 ? "07" : "01")	// North/South summer month if DST currently off
		var testdate = "" + year + "-" + month + "-01 " + ("0"+hour).slice(-2) + ":00:00"
		var testname = "hr$testdate"+now.valueOf()+"_"+FSO.GetTempName()
		var testfile = FSO.BuildPath(FSO.GetSpecialFolder(2),testname)
		var command = touch + ' -d "' + testdate + '" "' + testfile + '"'
		var shell = WSH.Exec(command)
		for (; shell.Status == 0; WScript.Sleep(10) );
		var fileobj  = FSO.GetFile(testfile)
		var filedate = new Date(Date.parse(fileobj.DateLastModified))
		var filetime = filedate.getHours()*60 + filedate.getMinutes()
		Date.applyFileDateCorrection = (Math.abs(tz2-tz1) == Math.abs(filetime-hour*60) )
		FSO.DeleteFile(testfile)
	}
	Date.checkFileDateFlaw()


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// END-OF-FILE  --  hr$wsh-fso.js  --  Copyright (c) 2005,2007 Henk Reints, http://henk-reints.nl
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
